from __future__ import annotations

import sys
from collections.abc import AsyncGenerator, Generator
from contextlib import asynccontextmanager, contextmanager
from enum import StrEnum
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Annotated, Literal, Self, cast
from uuid import uuid4

from anyio import TemporaryDirectory as AsyncTemporaryDirectory
from asyncer import asyncify
from datamodel_code_generator import (
    DataModelType,
    InputFileType,
    PythonVersion,
    generate,
)
from pydantic import BaseModel, Field, model_validator


class _Type(StrEnum):
    # Primitive types
    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    # Advanced types
    OBJECT = "object"
    ARRAY = "array"


class _BaseSchema(BaseModel):
    description: Annotated[str, Field(min_length=1)]


class String(_BaseSchema):
    """
    Pydantic model representing a JSON Schema string.
    """

    # default=_Type.STRING tells Pydantic to add the default to the type metadata and use it during instance creation/validation
    # = _Type.STRING saves the developer from having to specify the type manually during instance creation
    type: Annotated[Literal[_Type.STRING], Field(default=_Type.STRING)] = _Type.STRING


class Enum(String):
    """
    Pydantic model representing a JSON Schema enum, based on the string type.
    """

    enum: Annotated[set[str], Field(min_length=1)]


class Number(_BaseSchema):
    """
    Pydantic model representing a JSON Schema number.
    """

    type: Annotated[Literal[_Type.NUMBER], Field(default=_Type.NUMBER)] = _Type.NUMBER


class Integer(_BaseSchema):
    """
    Pydantic model representing a JSON Schema integer.
    """

    type: Annotated[Literal[_Type.INTEGER], Field(default=_Type.INTEGER)] = (
        _Type.INTEGER
    )


class Boolean(_BaseSchema):
    """
    Pydantic model representing a JSON Schema boolean.
    """

    type: Annotated[Literal[_Type.BOOLEAN], Field(default=_Type.BOOLEAN)] = (
        _Type.BOOLEAN
    )


class Object(_BaseSchema):
    """
    Pydantic model representing a JSON Schema object.
    """

    type: Annotated[Literal[_Type.OBJECT], Field(default=_Type.OBJECT)] = _Type.OBJECT
    properties: dict[
        str,
        String | Number | Integer | Boolean | Enum | Object | Array,
    ]
    required: set[str]

    @model_validator(mode="after")
    def check_required_keys_exist(self) -> Self:
        if not self.required.issubset(self.properties.keys()):
            raise ValueError(
                f"Required keys must be a subset of keys defined in properties. Required keys are {[*self.required]}; property keys are {[*self.properties.keys()]}"
            )
        return self


class Array(_BaseSchema):
    """
    Pydantic model representing a JSON Schema array.
    """

    type: Annotated[Literal[_Type.ARRAY], Field(default=_Type.ARRAY)] = _Type.ARRAY
    items: String | Number | Integer | Boolean | Enum | Object | Array


@contextmanager
def convert_json_schema_to_pydantic_model(
    object: Object, *, target_python_version: PythonVersion
) -> Generator[type[BaseModel]]:
    """
    Converts a JSON schema object into its Pydantic representation. This is useful while building language model-based applications which accept an arbitrary JSON representing the user's desired output structure.

    See corresponding tests in the `tests` directory for example usage.

    Args:
        object (Object): The JSON Schema object to convert. This should be an instance of the custom `Object` Pydantic model defined in this module, typically constructed to describe the desired schema for your data.
        target_python_version (PythonVersion): The Python version to target for the generated Pydantic model source code. Use values from `datamodel_code_generator.PythonVersion` (such as `PythonVersion.PY_311`) according to your environment.

    Returns:
        Generator[type[BaseModel]]: A context manager with the created Pydantic representation.
    """
    json_schema = object.model_dump_json(by_alias=True)

    with TemporaryDirectory() as temp_dir:
        # Create temporary destination for generated Python module containing the Pydantic model to generate
        temp_dir_path = Path(temp_dir)
        module_name = f"temp_pydantic_model_{uuid4().hex}"  # UUID to make sure no collision within sys.modules (see below)
        model_file = temp_dir_path / f"{module_name}.py"

        # Generate the Pydantic model from the JSON Schema object
        generate(
            json_schema,
            input_file_type=InputFileType.JsonSchema,
            output=model_file,
            output_model_type=DataModelType.PydanticV2BaseModel,
            target_python_version=target_python_version,
        )

        # Import the generated model
        spec = spec_from_file_location(module_name, model_file)
        if spec is None:  # pragma: no cover
            raise ValueError(
                "Spec for temporary module containing generated Pydantic model should not be None. Re-check how this function converts a JSON Schema to a Pydantic model."
            )
        if spec.loader is None:  # pragma: no cover
            raise ValueError(
                "Spec loader for temporary module containing generated Pydantic model should not be None. Re-check how this function converts a JSON Schema to a Pydantic model."
            )
        module = module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # Return the Pydantic model
        yield cast(type[BaseModel], module.Model)

        # Remove module from system when done. It will be deleted when the tempdir context exits
        sys.modules.pop(module_name)


@asynccontextmanager
async def convert_json_schema_to_pydantic_model_async(
    object: Object, *, target_python_version: PythonVersion
) -> AsyncGenerator[type[BaseModel]]:
    """
    Converts a JSON schema object into its Pydantic representation, async-style. This is useful while building language model-based applications which accept an arbitrary JSON representing the user's desired output structure.

    See corresponding tests in the `tests` directory for example usage.

    Args:
        object (Object): The JSON Schema object to convert. This should be an instance of the custom `Object` Pydantic model defined in this module, typically constructed to describe the desired schema for your data.
        target_python_version (PythonVersion): The Python version to target for the generated Pydantic model source code. Use values from `datamodel_code_generator.PythonVersion` (such as `PythonVersion.PY_311`) according to your environment.

    Returns:
        AsyncGenerator[type[BaseModel]]: An asynchronous context manager with the created Pydantic representation.
    """
    json_schema = object.model_dump_json(by_alias=True)

    async with AsyncTemporaryDirectory() as temp_dir:
        # Create temporary destination for generated Python module containing the Pydantic model to generate
        temp_dir_path = Path(temp_dir)
        module_name = f"temp_pydantic_model_{uuid4().hex}"  # UUID to make sure no collision within sys.modules (see below)
        model_file = temp_dir_path / f"{module_name}.py"

        # Generate the Pydantic model from the JSON Schema object
        await asyncify(generate)(
            json_schema,
            input_file_type=InputFileType.JsonSchema,
            output=model_file,
            output_model_type=DataModelType.PydanticV2BaseModel,
            target_python_version=target_python_version,
        )

        # Import the generated model
        spec = spec_from_file_location(module_name, model_file)
        if spec is None:  # pragma: no cover
            raise ValueError(
                "Spec for temporary module containing generated Pydantic model should not be None. Re-check how this function converts a JSON Schema to a Pydantic model."
            )
        if spec.loader is None:  # pragma: no cover
            raise ValueError(
                "Spec loader for temporary module containing generated Pydantic model should not be None. Re-check how this function converts a JSON Schema to a Pydantic model."
            )
        module = module_from_spec(spec)
        sys.modules[module_name] = module
        await asyncify(spec.loader.exec_module)(module)

        # Return the Pydantic model
        yield cast(type[BaseModel], module.Model)

        # Remove module from system when done. It will be deleted when the tempdir context exits
        sys.modules.pop(module_name)



from datetime import UTC, datetime


def utcdatetime(
    year: int,
    month: int,
    day: int,
    hour: int = 0,
    minute: int = 0,
    second: int = 0,
    microsecond: int = 0,
) -> datetime:
    """
    Generates a datetime object in UTC timezone.

    Args:
        year (int): Year.
        month (int): Month.
        day (int): Day.
        hour (int, optional): Hour. Defaults to 0.
        minute (int, optional): Minute. Defaults to 0.
        second (int, optional): Second. Defaults to 0.
        microsecond (int, optional): Microsecond. Defaults to 0.

    Returns:
        datetime: The datetime object in UTC.
    """
    return datetime(year, month, day, hour, minute, second, microsecond, tzinfo=UTC)



from collections.abc import Mapping, Sequence
from typing import Annotated, Any, Literal, TypeVar, cast, overload

from langchain.chat_models import init_chat_model
from langchain_community.callbacks.openai_info import (
    OpenAICallbackHandler,  # works with Google models, too
)
from langchain_core.callbacks.base import BaseCallbackManager
from langchain_core.language_models import LanguageModelInput
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableConfig
from pydantic import BaseModel, Field

_O = TypeVar("_O", bound=BaseModel)


class Output(BaseModel):
    content: Annotated[str, Field(description="Your response")]


class NoResult(Exception):
    pass


@overload
async def invoke_chat_model(
    model_id: str,
    prompt: str
    | Sequence[tuple[Literal["system", "human", "ai"], str | list[dict[str, str]]]]
    | ChatPromptTemplate,
    *,
    model_provider: str | None = None,
    configurable_fields: Literal["any"] | list[str] | tuple[str, ...] | None = None,
    config_prefix: str | None = None,
    model_kwargs: Mapping[str, Any] = dict(),
    output_structure: type[_O] = Output,
    output_structure_kwargs: Mapping[str, Any] = dict(),
    prompt_inputs: Mapping[str, Any] = dict(),
    use_default_callback_handler: Literal[True] = True,
    invoke_config: RunnableConfig | None = None,
    invoke_kwargs: Mapping[str, Any] = dict(),
) -> tuple[_O, OpenAICallbackHandler]: ...


@overload
async def invoke_chat_model(
    model_id: str,
    prompt: str
    | Sequence[tuple[Literal["system", "human", "ai"], str | list[dict[str, str]]]]
    | ChatPromptTemplate,
    *,
    model_provider: str | None = None,
    configurable_fields: Literal["any"] | list[str] | tuple[str, ...] | None = None,
    config_prefix: str | None = None,
    model_kwargs: Mapping[str, Any] = dict(),
    output_structure: type[_O] = Output,
    output_structure_kwargs: Mapping[str, Any] = dict(),
    prompt_inputs: Mapping[str, Any] = dict(),
    use_default_callback_handler: Literal[False] = False,
    invoke_config: RunnableConfig | None = None,
    invoke_kwargs: Mapping[str, Any] = dict(),
) -> _O: ...


async def invoke_chat_model(
    model_id: str,
    prompt: str
    | Sequence[tuple[Literal["system", "human", "ai"], str | list[dict[str, str]]]]
    | ChatPromptTemplate,
    *,
    model_provider: str | None = None,
    configurable_fields: Literal["any"] | list[str] | tuple[str, ...] | None = None,
    config_prefix: str | None = None,
    model_kwargs: Mapping[str, Any] = dict(),
    output_structure: type[_O] = Output,
    output_structure_kwargs: Mapping[str, Any] = dict(),
    prompt_inputs: Mapping[str, Any] = dict(),
    use_default_callback_handler: bool = True,
    invoke_config: RunnableConfig | None = None,
    invoke_kwargs: Mapping[str, Any] = dict(),
) -> _O | tuple[_O, OpenAICallbackHandler]:
    """
    Asynchronously calls an LLM in a single invocation with minimal requirements -- a model ID to use and a prompt.

    Make sure you've installed the appropriate LangChain package for the model provider you're using, e.g., `pip install langchain-google-vertexai` to access Google's Gemini models.

    Args:
        model_id (str): ID of the model to use, e.g., "gemini-2.0-flash-001".
        prompt (str | Sequence[tuple[Literal["system", "human", "ai"], str | list[dict[str, str]]]] | ChatPromptTemplate): The prompt to use. Can either be a string, a sequence of tuples fit to feed into `langchain.chat_models.ChatPromptTemplate`, or a `langchain.chat_models.ChatPromptTemplate` instance.
        model_provider (str | None): Model provider as required by `langchain.chat_models.init_chat_model`. Defaults to None.
        configurable_fields (Literal["any"] | list[str] | tuple[str, ...] | None): Configurable fields for the model, as required by `langchain.chat_models.init_chat_model`. Defaults to None.
        config_prefix (str | None): Configuration prefix for the model, as required by `langchain.chat_models.init_chat_model`. Defaults to None.
        model_kwargs (Mapping[str, Any]): Keyword arguments to pass `langchain.chat_models.init_chat_model`. Defaults to an empty dictionary. This would be, for example, the place to specify the desired model temperature and max output tokens.
        output_structure (type[_O]): Structure of the output. Defaults to Output.
        output_structure_kwargs (Mapping[str, Any]): Keyword arguments for the output structure. Defaults to an empty dictionary. This would be, for example, the place to specify the output structure method such as `json_mode`.
        prompt_inputs (Mapping[str, Any]): Inputs for the prompt during a LangChain `ainvoke` chain call, if the prompt is a LangChain template holding `{}` references to variables. Defaults to an empty dictionary.
        use_default_callback_handler (bool): Whether to use the default callback handler returned by this function. Defaults to True.
        invoke_config (RunnableConfig | None): Configuration for the LangChain `ainvoke` chain call. Defaults to None. This would, for example, contain LangChain callback handlers you'd like to pass to the model, in which case it's probably a good idea to set `use_default_callback_handler` to `False`.
        invoke_kwargs (Mapping[str, Any]): Keyword arguments for the invocation. Defaults to an empty dictionary.

    Returns:
        _O | tuple[_O, OpenAICallbackHandler]: Model output. If `use_default_callback_handler` is `True`, the output is a tuple containing the model output and the default callback handler. If `use_default_callback_handler` is `False`, the output is just the model output.

    Raises:
        NoResult: If the model call doesn't return anything, which typically happens when using a model with not enough max output tokens specified.
    """
    llm = cast(
        Runnable[LanguageModelInput, _O | None],
        init_chat_model(
            model=model_id,
            model_provider=model_provider,
            configurable_fields=configurable_fields,
            config_prefix=config_prefix,
            **model_kwargs,
        ).with_structured_output(output_structure, **output_structure_kwargs),
    )

    if isinstance(prompt, str):
        _prompt = ChatPromptTemplate([("human", prompt)])
    elif isinstance(prompt, Sequence):
        _prompt = ChatPromptTemplate(prompt)
    else:
        _prompt = prompt

    chain = _prompt | llm
    default_callback_handler = OpenAICallbackHandler()

    if use_default_callback_handler:
        if invoke_config is None:
            _invoke_config = RunnableConfig(callbacks=[default_callback_handler])
        elif (callbacks := invoke_config.get("callbacks")) is None:
            _invoke_config = RunnableConfig(
                {**invoke_config, "callbacks": [default_callback_handler]}
            )
        elif isinstance(callbacks, BaseCallbackManager):
            _invoke_config = RunnableConfig(
                {
                    **invoke_config,
                    "callbacks": callbacks.merge(
                        BaseCallbackManager(handlers=[default_callback_handler])
                    ),
                }
            )
        else:
            _invoke_config = RunnableConfig(
                {**invoke_config, "callbacks": [*callbacks, default_callback_handler]}
            )
    else:
        _invoke_config = invoke_config

    output = await chain.ainvoke({**prompt_inputs}, _invoke_config, **invoke_kwargs)

    if output is None:
        raise NoResult(
            "Chain output is None. This is likely due to using a model with not enough output tokens. Try using a different model or increasing the output token limit."
        )

    if use_default_callback_handler:
        return output, default_callback_handler

    return output



from base64 import urlsafe_b64encode
from collections.abc import Sequence
from email.message import EmailMessage

from asyncer import asyncify
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build


def send_html_email(
    subject: str,
    body: str,
    sender: str,
    recipients: Sequence[str],
    cc: Sequence[str],
    bcc: Sequence[str],
    *,
    token: Credentials,
) -> None:
    """
    Sends an HTML email using the Gmail API.

    Args:
        subject: Subject of the email.
        body: Body of the email.
        sender: Sender of the email.
        recipients: Recipients of the email.
        cc: Carbon copy recipients of the email.
        bcc: Blind carbon copy recipients of the email.
        token: OAuth2 credentials for the Gmail API.

    Returns:
        None
    """
    service = build("gmail", "v1", credentials=token)

    message = EmailMessage()
    message["Subject"] = subject
    message.set_content(body, subtype="html")

    message["From"] = sender
    # All recipients are managed through Google Groups
    message["To"] = recipients
    message["Cc"] = cc
    message["Bcc"] = bcc

    encoded_message = urlsafe_b64encode(message.as_bytes()).decode()

    # Send the message
    request = (
        service.users().messages().send(userId="me", body={"raw": encoded_message})
    )
    request.execute()


async def send_html_email_async(
    subject: str,
    body: str,
    sender: str,
    recipients: Sequence[str],
    cc: Sequence[str],
    bcc: Sequence[str],
    *,
    token: Credentials,
) -> None:
    """
    Sends an HTML email using the Gmail API asynchronously.

    Args:
        subject: Subject of the email.
        body: Body of the email.
        sender: Sender of the email.
        recipients: Recipients of the email.
        cc: Carbon copy recipients of the email.
        bcc: Blind carbon copy recipients of the email.
        token: OAuth2 credentials for the Gmail API.

    Returns:
        None
    """
    service = await asyncify(build)("gmail", "v1", credentials=token)

    message = EmailMessage()
    message["Subject"] = subject
    message.set_content(body, subtype="html")

    message["From"] = sender
    # All recipients are managed through Google Groups
    message["To"] = recipients
    message["Cc"] = cc
    message["Bcc"] = bcc

    encoded_message = urlsafe_b64encode(message.as_bytes()).decode()

    # Send the message
    request = (
        service.users().messages().send(userId="me", body={"raw": encoded_message})
    )
    await asyncify(request.execute)()



from datetime import UTC, datetime


def utcnow() -> datetime:
    """
    Generates a current timestamp in UTC timezone.

    Returns:
        datetime: The current timestamp in UTC.
    """
    return datetime.now(tz=UTC)



from collections.abc import Mapping
from datetime import datetime
from typing import TypedDict

from google.auth.transport._aiohttp_requests import Request as AsyncRequest
from google.auth.transport.requests import Request
from google.oauth2._service_account_async import (
    IDTokenCredentials as AsyncIDTokenCredentials,
)
from google.oauth2.service_account import IDTokenCredentials


class JwtClaims(TypedDict):
    # Issued at time (seconds since epoch)
    iat: int
    # Expiration time (seconds since epoch, default will be 1 hour from iat but can be set to a maximum of 12 hours)
    exp: int
    # Issuer (service account email, same as the email address in `credentials_info`)
    iss: str
    # Target audience (same as `target_audience`)
    target_audience: str


def get_service_account_oidc_jwt(
    credentials_info: Mapping[str, str],
    target_audience: str,
    *,
    additional_claims: JwtClaims | None = None,
    quota_project_id: str | None = None,
    universe_domain: str | None = None,
) -> tuple[str, datetime]:
    """
    Generates an OpenID Connect (OIDC) JSON Web Token (JWT) for a Google service account with respect to a target audience.
    This is primarily used for authenticating the service account to a third-party service such as an Cloud Run Service-run API.

    (For more information about this function's arguments, see: https://github.com/googleapis/google-auth-library-python/blob/c6d99030b8d972105913006bd052cf762b94a976/google/oauth2/service_account.py#L556.)

    Args:
        credentials_info (Mapping[str, str]): Credentials info for the service account. Normally, this is the JSON content of the service account key file.
        target_audience (str): Target audience for the JWT.
        additional_claims (Mapping[str, str] | None, optional): Additional claims to include in the JWT. Defaults to None. (For example, if you want to extend the JWT's lifetime, specify a higher value for the `exp` key -- maximum is 12 hours from `iat`.)
        quota_project_id (str | None, optional): Quota project ID. Defaults to None.
        universe_domain (str | None, optional): Universe domain. Defaults to None.

    Returns:
        tuple[str, datetime]: A tuple containing the JWT and its expiration time.
    """
    credentials = IDTokenCredentials.from_service_account_info(
        credentials_info,
        target_audience=target_audience,
        additional_claims=additional_claims,
        quota_project_id=quota_project_id,
        universe_domain=universe_domain,
    )
    credentials.refresh(Request())
    return str(credentials.token), credentials.expiry


async def get_service_account_oidc_jwt_async(
    credentials_info: Mapping[str, str],
    target_audience: str,
    *,
    additional_claims: JwtClaims | None = None,
    quota_project_id: str | None = None,
    universe_domain: str | None = None,
) -> tuple[str, datetime]:
    """
    Generates an OpenID Connect (OIDC) JSON Web Token (JWT) for a Google service account with respect to a target audience.
    This is primarily used for authenticating the service account to a third-party service such as an Cloud Run Service-run API.

    Args:
        credentials_info (Mapping[str, str]): Credentials info for the service account. Normally, this is the JSON content of the service account key file.
        target_audience (str): Target audience for the JWT.
        additional_claims (Mapping[str, str] | None, optional): Additional claims to include in the JWT. Defaults to None. (For example, if you want to extend the JWT's lifetime, specify a higher value for the `exp` key -- maximum is 12 hours from `iat`.)
        quota_project_id (str | None, optional): Quota project ID. Defaults to None.
        universe_domain (str | None, optional): Universe domain. Defaults to None.

    Returns:
        tuple[str, datetime]: A tuple containing the JWT and its expiration time.
    """
    credentials = AsyncIDTokenCredentials.from_service_account_info(
        credentials_info,
        target_audience=target_audience,
        additional_claims=additional_claims,
        quota_project_id=quota_project_id,
        universe_domain=universe_domain,
    )
    await credentials.refresh(AsyncRequest())
    return str(credentials.token), credentials.expiry



import inspect
from collections.abc import Iterable, Mapping
from datetime import date, datetime, time
from decimal import Decimal
from enum import StrEnum
from functools import partial, reduce
from types import NoneType, UnionType
from typing import Any, Literal, TypeAlias, Union, cast, get_args, get_origin

from annotated_types import MaxLen
from google.cloud.bigquery import SchemaField, StandardSqlTypeNames
from google.cloud.bigquery.schema import (
    _DEFAULT_VALUE,  # pyright: ignore[reportPrivateUsage]
)
from pydantic import BaseModel
from pydantic._internal._fields import PydanticMetadata


class Mode(StrEnum):
    REQUIRED = "REQUIRED"
    NULLABLE = "NULLABLE"
    REPEATED = "REPEATED"


_ClassInfo: TypeAlias = type[Any] | UnionType | tuple["_ClassInfo", ...]


def _class_is_subclass(
    obj: Any, base: _ClassInfo, *, excluded: _ClassInfo = tuple()
) -> bool:
    return (
        inspect.isclass(obj) and issubclass(obj, base) and not issubclass(obj, excluded)
    )


def convert_pydantic_model_to_schema(model: type[BaseModel]) -> list[SchemaField]:
    """
    Converts a Pydantic model representing a row in a BigQuery table to a list of SchemaField objects. Feature parity is not guaranteed, so use with caution.

    Args:
        model (BaseModel): The Pydantic model to convert.

    Returns:
        list[SchemaField]: A list of SchemaField objects representing the fields of the model, ready to be used in a BigQuery table.
    """
    schema: list[SchemaField] = []

    for field_name, field in model.model_fields.items():
        field_type = field.annotation
        mode: Mode = Mode.REQUIRED

        if (
            field_type is None
        ):  # pragma: no cover # Remove pragma: no cover once seeing this error in the wild
            raise NotImplementedError(
                f"Pydantic model field type of None is not supported. Got None for field {field_name}. Typically this should never be reached; if you're the developer of this library, please add a test case!"
            )

        # Try to strip None from union type and set mode to NULLABLE if None is indeed present
        if get_origin(field_type) in {UnionType, Union}:
            # Can use set here because X | int | int, for example, is the same as X | int
            union_type_args = set(get_args(field_type))
            if NoneType in union_type_args:
                mode = Mode.NULLABLE
            match list(union_type_args - {NoneType}):
                case [main_type]:
                    # Reassign field_type to be, e.g., X from X | None after setting mode to NULLABLE
                    field_type = main_type
                case _:
                    raise ValueError(
                        f"Pydantic model field type cannot be a union of more than one non-NoneType type; got {field_type} for field {field_name}"
                    )

        origin = get_origin(field_type)
        # Try to strip field type from list, set, or Iterable and set mode to REPEATED
        if _class_is_subclass(
            origin,
            tuple | list | set | frozenset | Iterable,
            # Mapping is a subclass of Iterable, but we want to treat it like a dictionary
            excluded=Mapping,
        ):
            mode = Mode.REPEATED
            # Use list here instead of set because Generator[int, None, int] is not the same as Generator[int]
            iterable_type_args = list(get_args(field_type))
            # Reassign field_type to be, e.g., X from list[X] after setting mode to REPEATED
            # iterable_type_args[0] will always exist, since if the type annotation is list instead of list[X], origin would be none, this code block would be skipped, and we'll go straight to  NotImplementedError: Unsupported field type: <class 'list'>. Even better, this usually would be caught by any half-decent IDE
            field_type = iterable_type_args[0]
            # Check if inside is a Union type with None. If so, strip None from it, same as above
            if get_origin(field_type) in {UnionType, Union}:
                union_type_args = set(get_args(field_type))
                match list(union_type_args - {NoneType}):
                    case [main_type]:
                        field_type = main_type
                    case _:
                        raise ValueError(
                            f"Pydantic model field type cannot contain a union of more than one non-NoneType type; got {field_type} for field {field_name}"
                        )

        # Try to parse literal values if the type is typing.Literal
        origin = get_origin(field_type)
        if origin is Literal:
            literal_args = get_args(field_type)
            literal_types = set(map(type, literal_args))
            if NoneType in literal_types and mode != Mode.REPEATED:
                mode = Mode.NULLABLE
            match literal_types_except_none := list(literal_types - {NoneType}):
                case [main_type]:
                    field_type = main_type
                case _:
                    raise ValueError(
                        f"Pydantic model field literal cannot contain values corresponding to more than one non-NoneType type; got {field_type} corresponding to {literal_types_except_none} for field {field_name}"
                    )

        _SchemaField = partial(
            SchemaField,
            name=field_name,
            description=field.description or _DEFAULT_VALUE,
            mode=mode,
        )

        field_metadata = field.metadata
        field_pydantic_metadata = reduce(
            lambda acc, obj: {**acc, **obj.__dict__},
            [obj for obj in field_metadata if isinstance(obj, PydanticMetadata)],
            dict(),
        )

        # Get origin again
        origin = get_origin(field_type)

        # Infer BigQuery field type from field_type and origin
        if _class_is_subclass(origin, dict | Mapping):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.JSON)

        elif _class_is_subclass(field_type, BaseModel):
            # Recursively convert the inner model to schema for the struct field
            field_schema = _SchemaField(
                field_type=StandardSqlTypeNames.STRUCT,
                fields=convert_pydantic_model_to_schema(field_type),
            )

        elif _class_is_subclass(field_type, bool):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.BOOL)

        elif _class_is_subclass(field_type, int):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.INT64)

        elif _class_is_subclass(field_type, float):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.FLOAT64)

        elif _class_is_subclass(field_type, Decimal):
            precision = cast(int, field_pydantic_metadata.get("max_digits", 0))
            scale = cast(int, field_pydantic_metadata.get("decimal_places", 0))
            if precision <= 38 and scale <= 9:
                field_schema = _SchemaField(
                    field_type=StandardSqlTypeNames.NUMERIC,
                    precision=precision,
                    scale=scale,
                )
            elif precision <= 76 and scale <= 38:
                field_schema = _SchemaField(
                    field_type=StandardSqlTypeNames.BIGNUMERIC,
                    precision=precision,
                    scale=scale,
                )
            else:
                raise ValueError(
                    f"Precision and scale values are out of range. Maximum precision possible for Decimal is 76 and maximum scale is 38. Got precision={precision} and scale={scale} for field {field_name}"
                )

        elif _class_is_subclass(field_type, str):
            max_length = _DEFAULT_VALUE
            for metadata in field_metadata:
                if isinstance(metadata, MaxLen):
                    max_length = metadata.max_length
            field_schema = _SchemaField(
                field_type=StandardSqlTypeNames.STRING, max_length=max_length
            )

        elif _class_is_subclass(field_type, bytes):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.BYTES)

        elif _class_is_subclass(field_type, datetime):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.TIMESTAMP)

        elif _class_is_subclass(field_type, date):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.DATE)

        elif _class_is_subclass(field_type, time):
            field_schema = _SchemaField(field_type=StandardSqlTypeNames.TIME)

        else:
            raise NotImplementedError(
                f"Unsupported field type; got {field_type} for {field_name}"
            )

        schema.append(field_schema)

    return schema



from __future__ import annotations

import asyncio
import time
from collections.abc import AsyncIterator, Iterator
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    asynccontextmanager,
    contextmanager,
)
from dataclasses import dataclass
from enum import IntEnum, auto
from typing import Generic, Literal, TypeVar, cast


class _TimerState(IntEnum):
    """
    Enum that represents the state of a timer.
    """

    STARTED = auto()
    STOPPED = auto()


_S = TypeVar("_S", bound=_TimerState)


@dataclass(slots=True)
class _Timer(Generic[_S]):
    """
    Class that times the execution of a block of code and stores seconds elapsed.
    Uses typestate pattern to discourage (but not prevent) it from being used more than once.
    """

    seconds_elapsed: float = 0

    @contextmanager
    def __call__(
        self: _Timer[Literal[_TimerState.STARTED]],
    ) -> Iterator[_Timer[Literal[_TimerState.STOPPED]]]:
        """
        Times the execution.

        Yields:
            Iterator[_Timer[Literal[_TimerState.STOPPED]]]: The timer instance.
        """
        start = time.perf_counter()
        try:
            yield cast(_Timer[Literal[_TimerState.STOPPED]], self)
        finally:
            end = time.perf_counter()
            self.seconds_elapsed = end - start


@dataclass(slots=True)
class _AsyncTimer(Generic[_S]):
    """
    Class that times the execution of an asynchronous block of code and stores seconds elapsed.
    Uses typestate pattern to discourage (but not prevent) it from being used more than once.
    """

    seconds_elapsed: float = 0

    @asynccontextmanager
    async def __call__(
        self: _AsyncTimer[Literal[_TimerState.STARTED]],
    ) -> AsyncIterator[_AsyncTimer[Literal[_TimerState.STOPPED]]]:
        """
        Times the execution.

        Yields:
            AsyncIterator[_AsyncTimer[Literal[_TimerState.STOPPED]]]: The timer instance.
        """
        loop = asyncio.get_event_loop()
        start = loop.time()
        try:
            yield cast(_AsyncTimer[Literal[_TimerState.STOPPED]], self)
        finally:
            end = loop.time()
            self.seconds_elapsed = end - start


def time_execution() -> AbstractContextManager[_Timer[Literal[_TimerState.STOPPED]]]:
    """
    Times the execution of a block of code and returns the Timer instance.
    """
    return _Timer[Literal[_TimerState.STARTED]]()()


def time_execution_async() -> AbstractAsyncContextManager[
    _AsyncTimer[Literal[_TimerState.STOPPED]]
]:
    """
    Times the execution of an asynchronous block of code and returns the AsyncTimer instance.
    """
    return _AsyncTimer[Literal[_TimerState.STARTED]]()()



from datetime import timedelta

from dagster import AutomationCondition
from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.declarative_automation import AndAutomationCondition


def _validate_lookback_range(
    start: timedelta | None, end: timedelta | None
) -> AutomationCondition[T_EntityKey]:
    match (start, end):
        case (None, timedelta()):
            raise ValueError(
                "If lookback_end is specified, lookback_start must also be specified."
            )
        case (timedelta(), timedelta()) if end.total_seconds() >= start.total_seconds():
            raise ValueError("lookback_start must precede lookback_end")
        case (timedelta(), timedelta()):
            window = AutomationCondition.in_latest_time_window(
                lookback_delta=start
            ) & ~AutomationCondition.in_latest_time_window(lookback_delta=end)
        case (timedelta(), None):
            window = AutomationCondition.in_latest_time_window(lookback_delta=start)
        case (None, None):
            window = AutomationCondition.in_latest_time_window()

    return window


def on_cron_persistent(
    cron_schedule: str,
    cron_timezone: str = "UTC",
    lookback_start: timedelta | None = None,
    lookback_end: timedelta | None = None,
) -> AndAutomationCondition[T_EntityKey]:
    """
    Returns an AutomationCondition that triggers asset execution on a given cron schedule until the asset is successfully materialized.

    If the asset is time-partitioned, unlike `AutomationCondition.on_cron`, multiple time partitions can be considered by setting `lookback_start` and `lookback_end` to the desired timedelta values (relative to the partition time).

    Args:
        cron_schedule (str): Cron schedule for the automation condition.
        cron_timezone (str): Timezone for the cron schedule. Defaults to "UTC".
        lookback_start (timedelta | None): Start of the lookback window. Defaults to None.
        lookback_end (timedelta | None): End of the lookback window. Defaults to None.

    Returns:
        AndAutomationCondition[T_EntityKey]: The returned AutomationCondition.
    """
    window = _validate_lookback_range(lookback_start, lookback_end)

    return (
        window
        & AutomationCondition.cron_tick_passed(
            cron_schedule, cron_timezone
        ).since_last_handled()
        & AutomationCondition.all_deps_updated_since_cron(cron_schedule, cron_timezone)
        & ~AutomationCondition.in_progress()
        & (AutomationCondition.missing() | AutomationCondition.execution_failed())
    ).with_label("on_cron_persistent")


def eager_persistent(
    lookback_start: timedelta | None = None, lookback_end: timedelta | None = None
) -> AndAutomationCondition[T_EntityKey]:
    """
    Returns an AutomationCondition which will cause a target to be executed if any of
    its dependencies update, and will execute missing partitions if they become missing
    after this condition is applied to the target.

    This will not execute targets that have any missing or in progress dependencies, or
    are currently in progress. But, unlike `AutomationCondition.eager`, it will also execute
    targets that just failed.

    For time partitioned assets, unlike `AutomationCondition.eager`, multiple time partitions can be considered by setting `lookback_start` and `lookback_end` to the desired timedelta values (relative to the partition time).

    Args:
        lookback_start (timedelta | None): Start of the lookback window. Defaults to None.
        lookback_end (timedelta | None): End of the lookback window. Defaults to None.

    Returns:
        AndAutomationCondition[T_EntityKey]: The returned AutomationCondition.
    """
    window = _validate_lookback_range(lookback_start, lookback_end)

    return (
        window
        & (
            AutomationCondition.missing().since_last_handled()
            | AutomationCondition.any_deps_updated().since_last_handled()
            | AutomationCondition.execution_failed()
        )
        & ~AutomationCondition.any_deps_missing()
        & ~AutomationCondition.any_deps_in_progress()
        & ~AutomationCondition.in_progress()
    ).with_label("eager_persistent")



from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from functools import wraps
from typing import Generic, ParamSpec, TypeVar

_P = ParamSpec("_P")
_R = TypeVar("_R")
_E = TypeVar("_E", bound=Exception)


@dataclass
class Success(Generic[_R]):
    """
    Wrapper class for successful function calls.
    """

    value: _R


class propagate_error(Generic[_E]):
    """
    Decorator that wraps a function and returns a Success instance if the function runs successfully, or an exception if it fails.
    The exception type is specified as a type argument, used to type hint the return value of the decorated function for downstream match statements. (See the corresponding test for usage.)
    """

    def __call__(
        self, func: Callable[_P, _R]
    ) -> Callable[_P, Success[_R] | _E | Exception]:
        """Inner decorator that wraps the function and returns a Success instance if the function runs successfully, or an exception if it fails.

        Args:
            func (Callable[_P, _R]): The function to be wrapped.

        Returns:
            Callable[_P, Success[_R] | _E | Exception]: The wrapped function.
        """

        @wraps(func)
        def wrapper(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> Success[_R] | _E | Exception:
            try:
                return Success(func(*args, **kwargs))
            except Exception as e:
                return e

        return wrapper


class propagate_error_async(Generic[_E]):
    """
    Error-propagation decorator for async functions.
    """

    def __call__(
        self, func: Callable[_P, Awaitable[_R]]
    ) -> Callable[_P, Awaitable[Success[_R] | _E | Exception]]:
        """
        Inner decorator that wraps the async function and returns a Success instance if the function runs successfully, or an exception if it fails.

        Args:
            func (Callable[_P, Awaitable[_R]]): The async function to be wrapped.

        Returns:
            Callable[_P, Awaitable[Success[_R] | _E | Exception]]: The wrapped async function.
        """

        @wraps(func)
        async def wrapper(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> Success[_R] | _E | Exception:
            try:
                return Success(await func(*args, **kwargs))
            except Exception as e:
                return e

        return wrapper
